﻿
Namespace XInput.Native

    Public Structure NativeGamepadState
        Implements IGamepadState

        Private Shared ReadOnly XnaButtons As New Dictionary(Of Button, ButtonStateBits) From _
        {
            {Button.A, ButtonStateBits.A},
            {Button.B, ButtonStateBits.B},
            {Button.X, ButtonStateBits.X},
            {Button.Y, ButtonStateBits.Y},
            {Button.LeftBumper, ButtonStateBits.Left_Shoulder},
            {Button.RightBumper, ButtonStateBits.Right_Shoulder},
            {Button.DPadUp, ButtonStateBits.DPad_Up},
            {Button.DPadDown, ButtonStateBits.DPad_Down},
            {Button.DPadLeft, ButtonStateBits.DPad_Left},
            {Button.DPadRight, ButtonStateBits.DPad_Right},
            {Button.Back, ButtonStateBits.Back},
            {Button.Start, ButtonStateBits.Start},
            {Button.LeftStickButton, ButtonStateBits.Left_Thumb},
            {Button.RightStickButton, ButtonStateBits.Right_Thumb}
        }

        Private Shared ReadOnly XnaPlayerIndices As New Dictionary(Of PlayerIndex, UInt32) From _
        {
            {PlayerIndex.One, 0},
            {PlayerIndex.Two, 1},
            {PlayerIndex.Three, 2},
            {PlayerIndex.Four, 3}
        }

        Private Const StickDeadzone As Integer = 8600
        Private Const TriggerDeadzone As Integer = 30

        Private _State As XINPUT_STATE
        Public ReadOnly Property State As XINPUT_STATE
            <DebuggerStepThrough()>
            Get
                Return _State
            End Get
        End Property

        Private _IsConnected As Boolean
        Public ReadOnly Property IsConnected As Boolean Implements IGamepadState.IsConnected
            <DebuggerStepThrough()>
            Get
                Return _IsConnected
            End Get
        End Property

        Public ReadOnly Property PacketNumber As Integer Implements IGamepadState.PacketNumber
            Get
                Dim U As IntUnion
                U.Value_UInt32_0 = State.PacketNumber
                Return U.Value_Int32_0
            End Get
        End Property

        Public Sub New(NewState As XINPUT_STATE, NewIsConnected As Boolean)
            _State = NewState
            _IsConnected = NewIsConnected
        End Sub

        Public Function GetButtonIsDown(Button As Button) As Boolean Implements IGamepadState.GetButtonIsDown
            Return (State.Buttons And XnaButtons(Button)) <> 0
        End Function

        Public Function GetStickValue(Stick As Stick) As Vector2D Implements IGamepadState.GetStickValue
            Dim RawX As Int16
            Dim RawY As Int16
            Select Case Stick
                Case XInput.Stick.Left
                    RawX = State.LeftStickX
                    RawY = State.LeftStickY
                Case XInput.Stick.Right
                    RawX = State.RightStickX
                    RawY = State.RightStickY
                Case Else
                    Throw New NopeException
            End Select
            Return New Vector2D(ApplyDeadzone(RawX, Int16.MaxValue, StickDeadzone), _
                                ApplyDeadzone(RawY, Int16.MaxValue, StickDeadzone))
        End Function

        Private Shared Function ApplyDeadzone(RawValue As Double, Maximum As Double, Deadzone As Double) As Double
            Dim AbsoluteRawValue = Math.Abs(RawValue)
            If AbsoluteRawValue < Deadzone Then
                Return 0
            End If
            Return Math.Sign(RawValue) * (AbsoluteRawValue - Deadzone) / (Maximum - Deadzone)
        End Function

        Public Function GetTriggerValue(Trigger As Trigger) As Double Implements IGamepadState.GetTriggerValue
            Dim RawValue As Byte
            Select Case Trigger
                Case XInput.Trigger.Left
                    RawValue = State.LeftTrigger
                Case XInput.Trigger.Right
                    RawValue = State.RightTrigger
                Case Else
                    Throw New NopeException
            End Select
            Return ApplyDeadzone(RawValue, Byte.MaxValue, TriggerDeadzone)
        End Function

        <System.Runtime.InteropServices.DllImport("XInput9_1_0.dll")>
        Private Shared Function XInputGetState(UserIndex As UInt32, ByRef State As XINPUT_STATE) As UInt32
        End Function

        Public Shared Function GetState(PlayerIndex As PlayerIndex) As NativeGamepadState
            Const ERROR_DEVICE_NOT_CONNECTED As UInt32 = 1167
            Const ERROR_SUCCESS As UInt32 = 0
            Dim State As New XINPUT_STATE
            Dim Result = XInputGetState(XnaPlayerIndices(PlayerIndex), State)
            If Result <> ERROR_SUCCESS AndAlso Result <> ERROR_DEVICE_NOT_CONNECTED Then
                Dim U As IntUnion
                U.Value_UInt32_0 = Result
                Throw New System.ComponentModel.Win32Exception(U.Value_Int32_0)
            End If
            Return New NativeGamepadState(State, Result <> ERROR_DEVICE_NOT_CONNECTED)
        End Function

    End Structure

End Namespace
